import _defineProperty from 'babel-runtime/helpers/defineProperty';
import _toConsumableArray from 'babel-runtime/helpers/toConsumableArray';
import _regeneratorRuntime from 'babel-runtime/regenerator';
import _Promise from 'babel-runtime/core-js/promise';
import _asyncToGenerator from 'babel-runtime/helpers/asyncToGenerator';
import _extends from 'babel-runtime/helpers/extends';
import _Object$keys from 'babel-runtime/core-js/object/keys';
import _slicedToArray from 'babel-runtime/helpers/slicedToArray';
import _Object$assign from 'babel-runtime/core-js/object/assign';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import _createClass from 'babel-runtime/helpers/createClass';
import { get, intersection, isEmpty, isEqual, isObject, mapValues, omit, omitBy, pull, transform, values } from 'lodash';
import _typeof from 'babel-runtime/helpers/typeof';
import lscache from 'lscache';
import _JSON$stringify from 'babel-runtime/core-js/json/stringify';

var Http = function () {
  function Http() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Http);

    _Object$assign(this, {
      config: {
        minCache: 5,
        host: '',
        basePath: ''
      }
    });
    this.storage = lscache;
    _Object$assign(this.config, config);
    if (typeof window.fetch !== 'function') {
      throw new Error('missing fetch');
    }
  }

  _createClass(Http, [{
    key: '_stripData',
    value: function _stripData(data, strip) {
      if (strip) {
        return get(data, strip);
      }
      return data;
    }
  }, {
    key: '_getKey',
    value: function _getKey(filename) {
      var _filename$split$slice = filename.split('/').slice(-1),
          _filename$split$slice2 = _slicedToArray(_filename$split$slice, 1),
          name = _filename$split$slice2[0];

      var _name$split = name.split('.'),
          _name$split2 = _slicedToArray(_name$split, 1),
          key = _name$split2[0];

      return key;
    }
  }, {
    key: 'get',
    value: function get$$1(filename) {
      var _this = this;

      var _ref = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {},
          _ref$update = _ref.update,
          update = _ref$update === undefined ? false : _ref$update,
          _ref$json = _ref.json,
          json = _ref$json === undefined ? false : _ref$json,
          _ref$latest = _ref.latest,
          latest = _ref$latest === undefined ? 0 : _ref$latest,
          _ref$updatedField = _ref.updatedField,
          updatedField = _ref$updatedField === undefined ? 'date' : _ref$updatedField;

      return new _Promise(function (resolve, reject) {
        var cached = _this.storage.get(filename);
        if (!update && cached) {
          var timestamp = get(cached, updatedField);
          // console.log('latest:update:cached 1', latest, update, !!cached, updatedField, timestamp);
          if (latest > timestamp) {
            update = true;
          }
          if (!update) {
            return resolve(_this._stripData(cached, json.strip));
          }
        }

        var arg = '';
        // console.log('latest:update:cached', latest, update, !!cached, get(cached, updatedField));
        if (update || !cached && latest === true) {
          arg = '?_=' + new Date().getTime();
        }
        var _config = _this.config,
            _config$host = _config.host,
            host = _config$host === undefined ? '' : _config$host,
            _config$basePath = _config.basePath,
            basePath = _config$basePath === undefined ? '' : _config$basePath;

        var baseUrl = '' + host + basePath;
        window.fetch('' + baseUrl + filename + arg).then(function (res) {
          return res.text();
        }).then(function (text) {
          if (!text) {
            throw new Error('no resp: ' + baseUrl + filename + arg);
          }
          try {
            return JSON.parse(text);
          } catch (e) {
            return text;
          }
        }).then(function (data) {
          var expireMins = void 0;
          if (latest === true) {
            expireMins = _this.config.minCache;
          }
          if ((typeof data === 'undefined' ? 'undefined' : _typeof(data)) === 'object') {
            _this.storage.set(filename, data, expireMins);
            resolve(_this._stripData(data, json.strip));
          } else {
            if (json.strip || json === true) {
              throw new Error('not valid json');
            } else {
              _this.storage.set(filename, data, expireMins);
              resolve(data);
            }
          }
        }).catch(function (err) {
          return reject(err);
        });
      });
    }
  }]);

  return Http;
}();

var storage = {};

var defaultPrefix = 'sto:';
var prefix = '';

function setPrefix(pre) {
  if (!prefix) {
    prefix = pre;
  } else {
    if (pre !== prefix) {
      throw new Error('prefix has already set as ' + prefix);
    }
  }
}

function set(key, json) {
  return new _Promise(function (resolve) {
    if (!prefix) {
      setPrefix(defaultPrefix);
    }
    if (json) {
      var str = _JSON$stringify(json);
      storage[key] = str;
      localStorage.setItem('' + prefix + key, str);
    } else {
      localStorage.removeItem('' + prefix + key);
    }
    resolve();
  });
}

function get$1(key) {
  return new _Promise(function (resolve, reject) {
    var stored = storage[key];
    if (!prefix) {
      setPrefix(defaultPrefix);
    }
    if (!stored) {
      stored = localStorage.getItem('' + prefix + key);
      if (stored) {
        storage[key] = stored;
      }
    }
    if (stored) {
      return resolve(JSON.parse(stored));
    }
    return reject('no data');
  });
}

function changes(obj, base, preserve) {
  return transform(obj, function (result, value, key) {
    var _preserve = Array.isArray(obj) ? { id: true, size: true } : null;
    if (!isEqual(value, base[key])) {
      result[key] = isObject(value) && isObject(base[key]) ? changes(value, base[key], _preserve) : value;
      if (preserve) {
        _Object$keys(preserve).forEach(function (preserveKey) {
          if (base[preserveKey]) {
            result[preserveKey] = base[preserveKey];
          }
        });
      }
    }
  });
}

var _config = {
  base: '/',
  ext: '.json',
  storage: {
    prefix: 'snap:'
  }
};

// TODO: split some functions to utils
// TODO: keep local list as a single list

var Snapmongo$1 = function () {
  function Snapmongo(collection) {
    var conf = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _classCallCheck(this, Snapmongo);

    setPrefix(_config.storage.prefix);
    _Object$assign(this, {
      schema: {
        fields: [],
        detailFields: []
      },
      name: '',
      isPending: false,
      isError: false,
      isReady: false,
      pendingResolves: [],
      pendingRejects: [],
      catalog: null,
      list: {},
      details: {},
      http: {},
      catalogName: 'index',
      config: {
        ext: _config.ext,
        base: _config.base
      },
      httpOptions: {
        json: true,
        timestampsField: 'meta.timestamps',
        updatedField: 'meta.updated'
      }
    });
    this.name = collection;
    _Object$assign(this.config, conf);
    this.http = new Http(this.config);
    this._requestCatalog().catch(function (err) {
      // console.error(err);
      throw err;
    });
    get$1('overrides:lists').then(function (res) {
      if (Array.isArray(res)) {
        var _res = _slicedToArray(res, 1),
            first = _res[0];

        set('overrides:lists', null);
        set('overrides:list', first);
      }
    }).catch();
  }

  _createClass(Snapmongo, [{
    key: '_getFilename',
    value: function _getFilename(name) {
      return this.config.base + '/' + this.name + '/' + name + this.config.ext;
    }
  }, {
    key: '_getLatest',
    value: function _getLatest(name) {
      var timestamps = get(this.catalog, this.httpOptions.timestampsField) || {};
      return timestamps[name];
    }
  }, {
    key: '_addList',
    value: function _addList(list) {
      var _this = this;

      _Object$keys(list).forEach(function (id) {
        _this.list[id] = _extends({}, list[id], { _id: id });
      });
    }
  }, {
    key: '_requestList',
    value: function _requestList(listName) {
      var _this2 = this;

      return this.http.get(this._getFilename(listName), _extends({}, this.httpOptions, {
        latest: this._getLatest(listName)
      })).then(function (data) {
        if (!data.list || !data.meta) {
          throw new Error('missing list and/or meta in the json');
        }
        _this2._addList(data.list);
        if (data.meta && data.meta.nextList) {
          return _this2._requestList(data.meta.nextList);
        }
      });
    }
  }, {
    key: '_determineConflicts',
    value: function () {
      var _ref = _asyncToGenerator(_regeneratorRuntime.mark(function _callee(remote) {
        var _ref2, _ref3, list, savedIds, catalogIds, conflictIds, save;

        return _regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return _Promise.all([get$1('overrides:list').catch(function () {
                  return {};
                }), get$1('overrides:details').catch(function () {
                  return {};
                })]);

              case 2:
                _ref2 = _context.sent;
                _ref3 = _slicedToArray(_ref2, 1);
                list = _ref3[0];
                savedIds = _Object$keys(list);
                catalogIds = _Object$keys(remote);
                conflictIds = intersection(savedIds, catalogIds);
                // console.log('------conflictIds', conflictIds);

                save = void 0;

                conflictIds.forEach(function (key) {
                  var diff = omit(changes(remote[key], list[key]), ['_id']);
                  if (isEmpty(diff)) {
                    save = true;
                    // console.log('identical!!!!', remote[key], list[key]);
                    // TODO: check details also
                    delete list[key];
                  } else {
                    // console.log('diff:', diff);
                  }
                });

                if (!save) {
                  _context.next = 12;
                  break;
                }

                return _context.abrupt('return', set('overrides:list', list));

              case 12:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function _determineConflicts(_x2) {
        return _ref.apply(this, arguments);
      }

      return _determineConflicts;
    }()
  }, {
    key: '_requestCatalog',
    value: function _requestCatalog() {
      var _this3 = this;

      if (this.isPending) {
        return new _Promise(function (res, rej) {
          _this3.pendingResolves.push(res);
          _this3.pendingRejects.push(rej);
        });
      }
      this.isPending = true;
      return this.http.get(this._getFilename(this.catalogName), _extends({}, this.httpOptions, {
        latest: true
      })).then(function () {
        var _ref4 = _asyncToGenerator(_regeneratorRuntime.mark(function _callee2(data) {
          var meta, listPromise;
          return _regeneratorRuntime.wrap(function _callee2$(_context2) {
            while (1) {
              switch (_context2.prev = _context2.next) {
                case 0:
                  if (!(!data.list || !data.meta)) {
                    _context2.next = 2;
                    break;
                  }

                  throw new Error('missing list and/or meta in the json');

                case 2:
                  _this3.catalog = data;
                  _context2.next = 5;
                  return _this3._determineConflicts(data.list);

                case 5:
                  _this3._addList(data.list);
                  _Object$assign(_this3.details, data.details);
                  meta = data.meta;

                  if (meta.fields) {
                    _Object$assign(_this3.schema.fields, meta.fields);
                  }
                  if (meta.detailFields) {
                    _Object$assign(_this3.schema.detailFields, meta.detailFields);
                  }
                  listPromise = _Promise.resolve();

                  if (_this3.catalog.meta.nextList) {
                    listPromise = _this3._requestList(_this3.catalog.meta.nextList);
                  }
                  return _context2.abrupt('return', listPromise.then(function () {
                    _this3.hasDetails = !!_this3.catalog.meta.detailFields;
                    _this3.isReady = true;
                    _this3.isError = false;
                    _this3.isPending = false;
                    _this3.pendingResolves.forEach(function (res) {
                      return res();
                    });
                    _this3.pendingResolves = [];
                    _this3.pendingRejects = [];
                  }));

                case 13:
                case 'end':
                  return _context2.stop();
              }
            }
          }, _callee2, _this3);
        }));

        return function (_x3) {
          return _ref4.apply(this, arguments);
        };
      }()).catch(function (err) {
        _this3.isError = true;
        _this3.isPending = false;
        _this3.pendingRejects.forEach(function (rej) {
          return rej(err);
        });
        _this3.pendingResolves = [];
        _this3.pendingRejects = [];
        throw err;
      });
    }
  }, {
    key: '_getBrief',
    value: function _getBrief(id) {
      var _this4 = this;

      var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      var brief = this.list[id];
      var localPromise = options.remoteOnly ? _Promise.resolve([{}]) : get$1('overrides:list').catch(function () {
        return {};
      });
      return localPromise.then(function (localList) {
        var localBrief = localList[id];
        if (!brief && !localBrief) {
          return null;
        }
        if (brief && _this4.hasDetails) {
          var detail = _this4.details[id];
          if (!detail && brief.hasDetail) {
            _Object$assign(brief, { _missing: true });
          } else {
            _Object$assign(brief, { _missing: false });
          }
        }
        return _Object$assign({}, brief, localBrief);
      });
    }
  }, {
    key: '_getData',
    value: function _getData(id, brief) {
      var _this5 = this;

      var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

      if (!brief.hasDetail) {
        return _Promise.resolve(_Object$assign(brief, { _id: id }));
      }
      var localPromise2 = options.remoteOnly ? _Promise.resolve({}) : get$1('overrides:details').catch(function () {
        return {};
      });
      return localPromise2.then(function (localDetails) {
        var localDetail = localDetails[id];
        var detail = _this5.details[id];
        return _Object$assign({}, brief, detail, localDetail, { _id: id });
      });
    }
  }, {
    key: '_getList',
    value: function _getList() {
      var _this6 = this;

      var _ref5 = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
          localList = _ref5.localList,
          _ref5$options = _ref5.options,
          options = _ref5$options === undefined ? {} : _ref5$options;

      var remote = this.list;
      var localPromise = options.remoteOnly ? _Promise.resolve({}) : localList ? _Promise.resolve(localList) : get$1('overrides:list').catch(function () {
        return {};
      });
      return localPromise.then(function (local) {
        var merged = omitBy(_Object$assign({}, remote, local), function (it) {
          return it._deleted;
        });
        // console.log('merged:', merged, values(merged).map(i=>i._id));
        var ret = values(merged);
        ret.reverse();
        ret.total = _this6.catalog.meta.total;
        return ret;
      });
    }
  }, {
    key: 'find',
    value: function find(q, options) {
      var _this7 = this;

      return this._requestCatalog().then(function () {
        return _this7._getList({ options: options });
      });
    }
  }, {
    key: 'findOne',
    value: function findOne(q, options) {
      var _this8 = this;

      var id = get(q, '_id') || q;
      var catalogPromise = _Promise.resolve();
      if (!this.isReady) {
        catalogPromise = this._requestCatalog();
      }
      return catalogPromise.then(_asyncToGenerator(_regeneratorRuntime.mark(function _callee3() {
        var brief;
        return _regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.next = 2;
                return _this8._getBrief(id, options);

              case 2:
                brief = _context3.sent;

                if (brief) {
                  _context3.next = 5;
                  break;
                }

                return _context3.abrupt('return', null);

              case 5:
                if (brief._missing) {
                  _context3.next = 7;
                  break;
                }

                return _context3.abrupt('return', _this8._getData(id, brief, options));

              case 7:
                return _context3.abrupt('return', _this8._requestData(brief).then(function () {
                  return _this8._getData(id, brief, options);
                }));

              case 8:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, _this8);
      })));
    }
  }, {
    key: '_requestData',
    value: function _requestData(brief) {
      var _this9 = this;

      return this.http.get(this._getFilename(brief._detail), _extends({}, this.httpOptions, {
        latest: this._getLatest(brief._detail)
      })).then(function (data) {
        _Object$assign(_this9.details, data.details);
      });
    }
  }, {
    key: '_addArray',
    value: function _addArray(target, src) {
      var _src = [].concat(src);
      pull(_src, target);
      _src.forEach(function (it) {
        return target.push(it);
      });
    }
  }, {
    key: '_getNewId',
    value: function _getNewId(localList) {
      return this._getList({ localList: localList }).then(function (list) {
        var max = list[0]._id;
        return '' + (parseInt(max, 10) + 1);
      });
    }
  }, {
    key: '_saveOverrides',
    value: function _saveOverrides(list, details) {
      return _Promise.all([set('overrides:list', list), set('overrides:details', details)]);
    }
  }, {
    key: '_updateSchema',
    value: function _updateSchema(fields, detailFields) {
      var schema = this.schema;

      this._addArray(schema.fields, fields);
      this._addArray(schema.detailFields, detailFields);
    }
  }, {
    key: '_splitDocument',
    value: function _splitDocument(doc) {
      var details = doc._details ? doc._details : {};
      var detailFields = _Object$keys(details);
      delete doc._details;
      var fields = _Object$keys(doc);
      this._addArray(detailFields, intersection(fields, this.schema.detailFields));
      pull.apply(undefined, [fields].concat(_toConsumableArray(detailFields)));
      return [doc, details, fields, detailFields];
    }
  }, {
    key: 'insert',
    value: function insert(obj) {
      var _this10 = this;

      return _Promise.all([get$1('overrides:list').catch(function () {
        return {};
      }), get$1('overrides:details').catch(function () {
        return {};
      })]).then(function (_ref7) {
        var _ref8 = _slicedToArray(_ref7, 2),
            list = _ref8[0],
            _details = _ref8[1];

        return _this10._getNewId(list).then(function (id) {
          var _splitDocument2 = _this10._splitDocument(_Object$assign({}, obj, { _id: id, _local: 'new' })),
              _splitDocument3 = _slicedToArray(_splitDocument2, 4),
              brief = _splitDocument3[0],
              details = _splitDocument3[1],
              fields = _splitDocument3[2],
              detailFields = _splitDocument3[3];

          _this10._updateSchema(fields, detailFields);
          _details = _extends({}, _details, _defineProperty({}, id, details));
          list[id] = brief;
          return _this10._saveOverrides(list, _details);
        });
      });
    }
  }, {
    key: 'update',
    value: function update(q, obj) {
      var _this11 = this;

      var id = '' + (get(q, '_id') || q);
      return _Promise.all([get$1('overrides:list').catch(function () {
        return {};
      }), get$1('overrides:details').catch(function () {
        return {};
      })]).then(function (_ref9) {
        var _ref10 = _slicedToArray(_ref9, 2),
            list = _ref10[0],
            _details = _ref10[1];

        // console.log(list, _details);
        var _splitDocument4 = _this11._splitDocument(_Object$assign({}, obj, { _id: id, _local: 'update' })),
            _splitDocument5 = _slicedToArray(_splitDocument4, 4),
            brief = _splitDocument5[0],
            details = _splitDocument5[1],
            fields = _splitDocument5[2],
            detailFields = _splitDocument5[3];

        _this11._updateSchema(fields, detailFields);
        _Object$assign(_details, _defineProperty({}, id, _this11.details[id]), _defineProperty({}, id, details));
        var remoteList = _this11.list;
        _Object$assign(list, remoteList && _defineProperty({}, id, remoteList[id]), _defineProperty({}, id, brief));
        // console.log('after:', list, _details);
        return _this11._saveOverrides(list, _details);
      });
    }
  }, {
    key: 'dumpUpdated',
    value: function dumpUpdated() {
      var _this12 = this;

      return _Promise.all([this._getList(), get$1('overrides:list').catch(function () {
        return {};
      }), get$1('overrides:details').catch(function () {
        return {};
      })]).then(function (_ref12) {
        var _ref13 = _slicedToArray(_ref12, 3),
            orgList = _ref13[0],
            localList = _ref13[1],
            localDetails = _ref13[2];

        if (_Object$keys(localList).length === 0) {
          return {};
        }
        var meta = _Object$assign({}, _this12.catalog.meta, {
          total: orgList.length,
          fields: _this12.schema.fields,
          detailFields: _this12.schema.detailFields
        });
        var list = _Object$assign({}, _this12.catalog.list, mapValues(localList, function (item) {
          return omit(item, '_local');
        }));
        var details = _Object$assign({}, _this12.catalog.details, localDetails);
        return {
          filename: 'index.json',
          data: {
            meta: meta,
            list: list,
            details: details
          }
        };
      });
    }
  }, {
    key: 'delete',
    value: function _delete(q) {
      var id = '' + (get(q, '_id') || q);
      return _Promise.all([get$1('overrides:list').catch(function () {
        return {};
      }), get$1('overrides:details').catch(function () {
        return {};
      })]).then(function (_ref14) {
        var _ref15 = _slicedToArray(_ref14, 1),
            list = _ref15[0];

        if (!list[id]) {
          list[id] = { _id: id };
        }
        list[id]._deleted = true;
        return set('overrides:list', list);
      });
    }
  }]);

  return Snapmongo;
}();

Snapmongo$1.config = function (_ref16) {
  var base = _ref16.base,
      ext = _ref16.ext,
      _ref16$storage = _ref16.storage;
  _ref16$storage = _ref16$storage === undefined ? {} : _ref16$storage;
  var prefix = _ref16$storage.prefix;

  _Object$assign(_config, {
    base: base,
    ext: ext
  });
  if (prefix) {
    _Object$assign(_config.storage, { prefix: prefix });
  }
};

export { Snapmongo$1 as Snapmongo };export default Snapmongo$1;
//# sourceMappingURL=index.es.js.map
