import { get } from 'lodash';
import lscache from 'lscache';

export class Http {
  constructor(config = {}) {
    Object.assign(this, {
      config: {
        minCache: 5,
        host: '',
        basePath: ''
      }
    });
    this.storage = lscache;
    Object.assign(this.config, config);
    if (typeof window.fetch !== 'function') {
      throw new Error('missing fetch');
    }
  }

  _stripData(data, strip) {
    if (strip) {
      return get(data, strip);
    }
    return data;
  }

  _getKey(filename) {
    const [name] = filename.split('/').slice(-1);
    const [key] = name.split('.');
    return key;
  }

  get(filename, {
    update = false,
    json = false,
    latest = 0,
    updatedField = 'date'
  } = {}) {
    return new Promise((resolve, reject) => {
      let cached = this.storage.get(filename);
      if (!update && cached) {
        const timestamp = get(cached, updatedField);
        // console.log('latest:update:cached 1', latest, update, !!cached, updatedField, timestamp);
        if (latest > timestamp) {
          update = true;
        }
        if (!update) {
          return resolve(this._stripData(cached, json.strip));
        }
      }

      let arg = '';
      // console.log('latest:update:cached', latest, update, !!cached, get(cached, updatedField));
      if (update || (!cached && latest === true)) {
        arg = '?_=' + new Date().getTime();
      }
      const {
        host = '',
        basePath = '',
      } = this.config;
      const baseUrl = `${ host }${ basePath }`;
      window.fetch(`${ baseUrl }${ filename }${ arg }`)
        .then(res => res.text())
        .then(text => {
          if (!text) {
            throw new Error(`no resp: ${baseUrl}${filename}${arg}`);
          }
          try {
            return JSON.parse(text);
          } catch (e) {
            return text;
          }
        })
        .then(data => {
          let expireMins;
          if (latest === true) {
            expireMins = this.config.minCache;
          }
          if (typeof data === 'object') {
            this.storage.set(filename, data, expireMins);
            resolve(this._stripData(data, json.strip));
          } else {
            if (json.strip || json === true) {
              throw new Error('not valid json');
            } else {
              this.storage.set(filename, data, expireMins);
              resolve(data);
            }
          }
        })
        .catch(err => reject(err));
    });
  }
}
export default Http;
