const storage = {};

const defaultPrefix = 'sto:';
let prefix = '';

export function setPrefix(pre) {
  if (!prefix) {
    prefix = pre;
  } else {
    if (pre !== prefix) {
      throw new Error('prefix has already set as ' + prefix);
    }
  }
}

export function set(key, json) {
  return new Promise((resolve) => {
    if (!prefix) {
      setPrefix(defaultPrefix);
    }
    if (json) {
      const str = JSON.stringify(json);
      storage[key] = str;
      localStorage.setItem(`${ prefix }${ key }`, str);
    } else {
      localStorage.removeItem(`${ prefix }${ key }`);
    }
    resolve();
  });
}

export function get(key) {
  return new Promise((resolve, reject) => {
    let stored = storage[key];
    if (!prefix) {
      setPrefix(defaultPrefix);
    }
    if (!stored) {
      stored = localStorage.getItem(`${ prefix }${ key }`);
      if (stored) {
        storage[key] = stored;
      }
    }
    if (stored) {
      return resolve(JSON.parse(stored));
    }
    return reject('no data');
  });
}

export default {
  setPrefix,
  set,
  get
};
